from huggingface_hub import hf_hub_download
from typing import Literal, List

AVAILABLE_FORMATS = [
    "Q2_K",
    "Q3_K_S",
    "Q3_K_M",
    "Q3_K_L",
    "Q4_K_S",
    "Q4_K_M",
    "Q5_K_S",
    "Q5_K_M",
    "Q6_K"
]

PROMPTING_STYLES: List[str] = [
    "Llama2",
    "OpenChat"
]


def prepare_model_to_load(
        model_name: str,
        quantize_format: str,
        provider: str = "erfanzar",
):
    """
    The prepare_model_to_load function is used to prepare the model name and format for loading.

    :param model_name: str: Specify the model name
    :param quantize_format: str: Specify the format of the model
    :param provider: str: Specify the provider of the model
    :return:  name and the file path of the quantized model
    """
    file_path = f"{model_name}.{quantize_format}.gguf"
    repo_id = f"{provider}/{model_name}"
    return repo_id, file_path


def download_model_gguf(
        pretrained_model_name_or_path: str | None = None,
        quantize_type: str | Literal[
            "Q2_K",
            "Q3_K_S",
            "Q3_K_M",
            "Q3_K_L",
            "Q4_K_S",
            "Q4_K_M",
            "Q5_K_S",
            "Q5_K_M",
            "Q6_K"
        ] = "Q4_K_S",
        model_name: str | Literal["Pixely1B-GGUF", "Pixely7B-GGUF"] | None = None,
        provider: str | Literal["erfanzar", "LucidBrains"] | None = None,
        hf_token: str = None
):
    if pretrained_model_name_or_path is not None and (model_name is None or provider is None):
        provider, model_name = pretrained_model_name_or_path.split("/")
    repo_id, file_path = prepare_model_to_load(
        model_name,
        quantize_type,
        provider
    )
    tkn = {}
    if hf_token is not None or hf_token != "":
        tkn = dict(token=hf_token)
    ref = hf_hub_download(
        repo_id=repo_id,
        filename=file_path,
        **tkn
    )
    return ref



from .inference import GGMLModel
from ..user_interface import GradioUserInference
import logging
from typing import List, Literal, Any
from ..utils import BaseClassAgent, DEFAULT_SYSTEM_PROMPT, prompt_model

logging.basicConfig(
    level=logging.INFO
)


class PixelyAIServeGGML(GradioUserInference, BaseClassAgent):
    def __init__(
            self,
            model: GGMLModel,
    ):
        self.model = model
        self.gradio_app_custom = self.create_gradio_pixely_ai()

    def process_gradio(
            self,
            prompt: str,
            history: List[List[str]],
            system_prompt: str | None,
            mode: str,
            max_length: int,
            max_new_tokens: int,
            max_compile_tokens: int,
            greedy: bool,
            temperature: float,
            top_p: float,
            top_k: int
    ):
        for holder, response in self.process(
                prompt=prompt,
                history=history,
                system_prompt=system_prompt,
                max_tokens=max_new_tokens,
                temperature=temperature,
                top_k=top_k,
                top_p=top_p,
                prompting_style="Llama2",
        ):
            yield holder, response

    @classmethod
    def from_pretrained(
            cls,
            pretrained_model_name_or_path: str,
            quantize_type: Literal[
                "Q2_K",
                "Q3_K_S",
                "Q3_K_M",
                "Q3_K_L",
                "Q4_K_S",
                "Q4_K_M",
                "Q5_K_S",
                "Q5_K_M",
                "Q6_K"
            ] = "Q4_K_S"
    ):
        return cls(
            GGMLModel.from_pretrained(
                pretrained_model_name_or_path=pretrained_model_name_or_path,
                quantize_type=quantize_type
            )
        )

    def process(
            self,
            prompt: str,
            history: List[List[str]],
            system_prompt: str,
            prompting_style: str | Literal["Llama2", "OpenChat"],
            max_tokens: int = 4096,
            temperature: float = 0.8,
            top_p: float = 0.9,
            top_k: int = 50
    ):
        """
        The sample function is the main entry point for a user to interact with the model.
        It takes in a prompt, which can be any string, and returns an iterator over
        strings that are generated by the model.
        The sample function also takes in some optional arguments:


        :param self: Refer to the current object
        :param prompt: str: Pass in the text that you want to generate a response for
        :param history: List[List[str]]: Keep track of the conversation history
        :param system_prompt: prompt for system 
        :param prompting_style: str: Prompting style to prompt Model
        :param max_tokens: int: Limit the number of tokens in a response
        :param temperature: float: Control the randomness of the generated text
        :param top_p: float: Control the probability of sampling from the top k tokens
        :param top_k: int: Control the number of candidates that are considered for each token
        :return: A generator that yields the next token in the sequence
        """
        template = self.model.get_chat_template(
            prompting_style
        )
        if self.model is not None:
            string = template(
                prompt,
                history,
                None if system_prompt == "" else system_prompt
            )
            history.append([prompt, ""])
            total_response = ""
            for response in self.model(
                    string,
                    top_k=top_k,
                    top_p=top_p,
                    max_position_embedding=max_tokens,
                    temperature=temperature,
            ):
                total_response += response.predictions.text
                history[-1][-1] = total_response
                yield "", history
        else:
            return [
                [prompt, "Opps Seems like you forgot to load me first ;\\"]
            ]

    def process_gradio_custom(self, prompt, user_id, data, system, max_new_tokens, greedy):
        history = data.split("<|END_OF_MESSAGE|>") if data != "" else []
        system = system if system != "" else DEFAULT_SYSTEM_PROMPT
        his = []
        for hs in history:
            if hs != "":
                his.append(hs.split("<|END_OF_MESSAGE_TURN_HUMAN|>"))
        history = his

        for holder, response in self.process(
                prompt=prompt,
                history=history,
                system_prompt=system,
                max_tokens=max_new_tokens,
                prompting_style="Llama2",
                top_p=0.95,
                top_k=50,
                temperature=0.8,
        ):
            yield holder, response[-1][-1]

    def launch(
            self,
            launch_custom: bool = True,
            launch_chat: bool = True,
            share: bool = True,
            *args,
            **kwargs
    ):
        urls = {}
        if launch_custom:
            self.gradio_app_custom.launch(share)
            urls["custom"] = self.gradio_app_custom.share_url
        if launch_chat:
            grad_chat = self.build_inference(
                self.process_gradio, 2048, 2048, 1
            )
            grad_chat.launch(share)
            urls["chat"] = grad_chat.share_url
        return urls

    @staticmethod
    def format_chat(history: List[List[str]], prompt: str, system: str = None) -> str:
        return prompt_model(message=prompt, system_prompt=system, chat_history=history)

    @staticmethod
    def prompt_model(message: str, chat_history, system_prompt: str):
        return prompt_model(message=message, chat_history=chat_history, system_prompt=system_prompt)



import jax.lax
from EasyDel import JAXServer, AutoEasyDelModelForCausalLM
from fjformer import get_float_dtype_by_name
from transformers import AutoTokenizer
import logging
from typing import List
from ..utils import BaseClassAgent, DEFAULT_SYSTEM_PROMPT, prompt_model

logging.basicConfig(
    level=logging.INFO
)


class PixelyAIServeJax(JAXServer, BaseClassAgent):
    def __init__(self, config=None):
        super().__init__(config=config)
        self.gradio_app_custom = self.create_gradio_pixely_ai()

    @classmethod
    def load_from_torch(cls, pretrained_model_name_or_path, config=None):

        tokenizer = AutoTokenizer.from_pretrained(pretrained_model_name_or_path)
        model, param = AutoEasyDelModelForCausalLM.from_pretrained(
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            dtype=get_float_dtype_by_name(config["dtype"] if config is not None else "fp16"),
            param_dtype=get_float_dtype_by_name(config["dtype"] if config is not None else "fp16"),
            precision=jax.lax.Precision("fastest"),
            device=jax.devices("cpu")[0]
        )
        return cls.load_from_params(
            config_model=model.config,
            model=model,
            config=config,
            params=param,
            tokenizer=tokenizer,
            add_params_field=True,
            do_memory_log=False
        )

    @classmethod
    def load_from_jax(cls, pretrained_model_name_or_path, checkpoint_path, config_repo=None, config=None):
        raise NotImplemented("Not Implemented Error!")

    def process_gradio_custom(self, prompt, user_id, data, system, max_new_tokens, greedy):
        history = data.split("<|END_OF_MESSAGE|>") if data != "" else []
        system = system if system != "" else DEFAULT_SYSTEM_PROMPT
        his = []
        for hs in history:
            if hs != "":
                his.append(hs.split("<|END_OF_MESSAGE_TURN_HUMAN|>"))
        history = his
        string = self.prompt_model(
            message=prompt,
            chat_history=history or [],
            system_prompt=system
        )
        response = "null"
        if not self.config.stream_tokens_for_gradio:

            for response, _ in self.process(
                    string=string,
                    greedy=greedy,
                    max_new_tokens=max_new_tokens,
            ):
                ...

        else:
            for response, _ in self.process(
                    string=string,
                    greedy=greedy,
                    max_new_tokens=max_new_tokens
            ):
                yield "", response
        return "", response

    @staticmethod
    def format_chat(history: List[List[str]], prompt: str, system: str = None) -> str:
        return prompt_model(message=prompt, system_prompt=system, chat_history=history)

    @staticmethod
    def prompt_model(message: str, chat_history, system_prompt: str):
        return prompt_model(message=message, chat_history=chat_history, system_prompt=system_prompt)



try:
    from .jax_serve import PixelyAIServeJax
except ModuleNotFoundError:
    print("\033[1;32mCouldn't import JAXServe   [IGNORE THIS IF YOU ARE NOT ON AI SIDE]\033[1;0m")

try:
    from .torch_serve import PixelyAIServeTorch
except ModuleNotFoundError:
    print("\033[1;32mCouldn't import TorchServe [IGNORE THIS IF YOU ARE NOT ON AI SIDE]\033[1;0m")

try:
    from .ggml_serve import PixelyAIServeGGML
except ModuleNotFoundError:
    print("\033[1;32mCouldn't import GGMLServe  [IGNORE THIS IF YOU ARE NOT ON AI SIDE]\033[1;0m")



from setuptools import setup, find_packages

setup(
    name="pixelyai_core",
    version="0.0.1",
    author="Erfan Zare Chavoshi",
    author_email="erfanzare82@eyahoo.com",
    description="Serve Utilises of PixelyAI",
    url="https://github.com/erfanzar/PixelyAI",
    packages=find_packages("src/python"),
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    keywords="machine learning, deep learning, pytorch, jax, flax, ggml, c++",
    install_requires=[
        "typing~=3.7.4.3",
        "ipython~=8.17.2",
        "tqdm~=4.64.1",
        "pydantic==2.5.3",
        "setuptools~=68.1.2",
        "gradio~=4.12.0",
        "uvicorn~=0.23.2",
        "fastapi~=0.104.1",
        "pydantic-core==2.14.6",
    ],
    python_requires=">=3.8",
    package_dir={"": "src/python"},
    password="PixelyAICore"
)



from llama_cpp import Llama
from dataclasses import dataclass
from typing import Optional, Union, List, Iterable, Literal
from ._setting import download_model_gguf


@dataclass
class InferencePredictions:
    text: str
    index: int
    logprobs: Optional[float]
    finish_reason: Optional[bool]


@dataclass
class InferenceOutput:
    id: str
    object: str
    created: str
    model: str
    predictions: InferencePredictions


class GGMLModel:
    def __init__(
            self,
            model: Llama,
            max_position_embedding: int = 2048
    ):
        """
        The __init__ function is the constructor for a class. It's called when an object of that class is created.

        :param self: Represent the instance of the class
        :param model: Llama: Pass the model object to the class
        :param max_position_embedding: int: Set the maximum number of tokens that can be used in a single query
        """
        self.model = model
        self.max_position_embedding = max_position_embedding

    @classmethod
    def from_pretrained(
            cls,
            pretrained_model_name_or_path: str,
            quantize_type: Literal[
                "Q2_K",
                "Q3_K_S",
                "Q3_K_M",
                "Q3_K_L",
                "Q4_K_S",
                "Q4_K_M",
                "Q5_K_S",
                "Q5_K_M",
                "Q6_K"
            ] = "Q4_K_S",
            max_position_embedding: int = 2048

    ):
        ref = download_model_gguf(
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            quantize_type=quantize_type
        )

        model = Llama(model_path=ref, n_ctx=max_position_embedding)
        return cls(
            model,
            max_position_embedding
        )

    def __call__(
            self,
            string: str,
            suffix: Optional[str] = None,
            max_position_embedding: Optional[int] = None,
            temperature: float = 0.8,
            top_p: float = 0.95,
            min_p: float = 0.05,
            typical_p: float = 1.0,
            logprobs: Optional[int] = None,
            echo: bool = False,
            stop=None,
            frequency_penalty: float = 0.0,
            presence_penalty: float = 0.0,
            repeat_penalty: float = 1.1,
            top_k: int = 40,
            seed: Optional[int] = None,
            tfs_z: float = 1.0,
            mirostat_mode: int = 0,
            mirostat_tau: float = 5.0,
            mirostat_eta: float = 0.1,
    ) -> Iterable[InferenceOutput]:
        """
       The __call__ function is the main function of this class.
       It takes in a string, and returns an iterable of InferenceOutput objects.
       The InferenceOutput object contains the following attributes:

       :param self: Bind the method to an object
       :param string: str: Pass the text to be completed by the model
       :param suffix: Optional[str]: Add a suffix to the end of the string
       :param max_position_embedding: Optional[int]: Limit the number of tokens that can be generated
       :param temperature: float: Control the randomness of the output
       :param top_p: float: Filter out the top p% of tokens
       :param min_p: float: Filter out tokens that have a probability less than min_p
       :param typical_p: float: Set the probability of a token being generated
       :param logprobs: Optional[int]: Control the number of log probabilities that are returned
       :param echo: bool: Determine whether to echo the string back as a prediction
       :param stop: Stop the inference at a certain token
       :param frequency_penalty: float: Penalize the frequency of words in the text
       :param presence_penalty: float: Penalize the presence of certain words in the output
       :param repeat_penalty: float: Penalize the model for repeating words
       :param top_k: int: Control the number of tokens that are considered for each step
       :param seed: Optional[int]: Set the seed for the random number generator
       :param tfs_z: float: Control the amount of text that is generated
       :param mirostat_mode: int: Determine which type of mirostat to use
       :param mirostat_tau: float: Control the rate of change in the probability distribution
       :param mirostat_eta: float: Control the amount of randomness in the model
       :param : Set the maximum number of tokens to generate
       :return: An iterable of inference-output objects
       """

        if stop is None:
            stop = []
        for model_response in self.model(
                string,
                stream=True,
                seed=seed,
                max_tokens=max_position_embedding or self.max_position_embedding,
                frequency_penalty=frequency_penalty,
                presence_penalty=presence_penalty,
                repeat_penalty=repeat_penalty,
                mirostat_mode=mirostat_mode,
                mirostat_tau=mirostat_tau,
                mirostat_eta=mirostat_eta,
                top_k=top_k,
                top_p=top_p,
                suffix=suffix,
                min_p=min_p,
                temperature=temperature,
                echo=echo,
                stop=stop,
                typical_p=typical_p,
                logprobs=logprobs
        ):
            predictions = InferencePredictions(
                **model_response["choices"][0]
            )
            predictions.text = predictions.text.replace("<0x0A>", "\n")
            response = InferenceOutput(
                predictions=predictions,
                created=model_response["created"],
                model=model_response["model"],
                object=model_response["object"],
                id=model_response["id"]
            )
            yield response

    @staticmethod
    def chat_template(
            message: str,
            chat_history: Optional[List[str] | List[List[str]]] = None,
            system_prompt: str = None
    ):
        """
        The chat_template function takes in a message, chat_history and system prompt.
        It then formats the message into a template that can be used to train the model.
        The function returns a string of text formatted as follows:

        :param message: str: Pass in the user's message to be added to the chat history
        :param chat_history: Optional[List[str] | List[List[str]]]: Pass in a list of strings or a list of lists
        :param system_prompt: str: Set the prompt for the system
        :return: prompt string
        """
        if chat_history is None:
            chat_history = []
        do_strip = False
        texts = [
            f'<s>[INST] <<SYS>>\n{system_prompt}\n<</SYS>>\n\n'
        ] if system_prompt is not None else [f'<s>[INST] ']
        for user_input, response in chat_history:
            user_input = user_input.strip() if do_strip else user_input
            do_strip = True
            texts.append(
                f'{user_input} [/INST] {response.strip()} </s><s>[INST] ')
        message = message.strip() if do_strip else message
        texts.append(f'{message} [/INST]')
        return ''.join(texts)

    @staticmethod
    def os_chat_template(
            message: str,
            chat_history: Optional[List[str] | List[List[str]]] = None,
            system_prompt: Optional[str] = None
    ):
        """
        The os_chat_template function takes in a message, chat history, and system prompt.
        It returns a string that is formatted to be used as the input for the OpenSubtitles dataset.
        The format of this string is:

        :param message: str: Pass in the user's message to the assistant
        :param chat_history: Optional[List[str] | List[List[str]]]: Specify the history of the conversation
        :param system_prompt: Optional[str]: Add a system prompt to the chat history
        :return: prompt string
        """
        if chat_history is None:
            chat_history = []
        system = f"<|system|>\n{system_prompt}</s>\n" if system_prompt is not None else ""
        ua = ""
        for user_input, response in chat_history:
            ua += f"<|user|>\n{user_input}</s>\n<|assistant|>\n{response}</s>\n"
        return system + ua + f"<|user|>\n{message}</s>\n<|assistant|>\n"

    def get_chat_template(self, template_name: Literal["Llama2", "OpenChat"] = "Llama2"):
        if template_name == "Llama2":
            return self.chat_template
        elif template_name == "OpenChat":
            return self.os_chat_template
        else:
            raise ValueError("UnKnown Chat Template requested")
